//
//  WiiMidiDelegate.m
//  WiiToMidi
//
//  Created by Mike Verdone on 22/12/06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "WiiMidiDelegate.h"

@implementation WiiMidiDelegate

-(id) initWithController:(id)c wiiId:(int)w wiiRemote:(WiiRemote*)wii {
	_controller = c;
	_wiiId = w;
	_wii = wii;
	_firstTime = 1;
	[self setLearnMode:FALSE];
	[self setSensitivity:0];
	
	MIDIClientCreate(CFSTR("WiiToMidi"), NULL, NULL, &_midi);
	MIDIOutputPortCreate(_midi, CFSTR("Output"), &_midiOut);
  
	int n = MIDIGetNumberOfDestinations();
	if (n > 0)
		_midiDest = MIDIGetDestination(0);
	if (_midiDest != NULL)
	{
		NSLog(@"Got MIDI destination.");
	}
	else
	{
		NSLog(@"No destinations :-(");
	}
	
	[_wii setDelegate:self];
	[_wii setMotionSensorEnabled:true];
	[_wii setLEDEnabled1:true enabled2:false enabled3:false enabled4:false];

	return self;
}

-(void) setMidiChannel:(int)midiChannel {
	_midiChannel = midiChannel;
}

// WiiRemoteDelegate messages:
-(void)dataChanged:(unsigned short)buttonData accX:(unsigned char)accX accY:(unsigned char)accY accZ:(unsigned char)accZ mouseX:(float)mx mouseY:(float)my
{
	int i;
	unsigned char signals[SIGNAL_ARR_SIZE];
	signals[SA_ACC_X] = accX;
	signals[SA_ACC_Y] = accY;
	signals[SA_ACC_Z] = accZ;
	signals[SA_BUT_A] = BUTTONDOWN(kWiiRemoteAButton, buttonData);
	signals[SA_BUT_B] = BUTTONDOWN(kWiiRemoteBButton, buttonData);
	signals[SA_BUT_1] = BUTTONDOWN(kWiiRemoteOneButton, buttonData);	
	signals[SA_BUT_2] = BUTTONDOWN(kWiiRemoteTwoButton, buttonData);	
	signals[SA_BUT_PLUS] = BUTTONDOWN(kWiiRemotePlusButton, buttonData);	
	signals[SA_BUT_MINUS] = BUTTONDOWN(kWiiRemoteMinusButton, buttonData);	

	if (_firstTime)
	{
		memcpy(_oldSignals, signals, SIGNAL_ARR_SIZE);
		_firstTime = FALSE;
		return;
	}

	// ButtonOne on buttonUp toggle learn mode.
	if ((! signals[SA_BUT_1]) && (_oldSignals[SA_BUT_1]))
	{
		[self setLearnMode:(!_learnMode)];
	}

	// ButtonPlus on buttonUp
	if ((! signals[SA_BUT_PLUS]) && (_oldSignals[SA_BUT_PLUS]))
	{
		[self setSensitivity:(_sensitivity + 1)];
	}

	// ButtonMinus on buttonUp
	if ((! signals[SA_BUT_MINUS]) && (_oldSignals[SA_BUT_MINUS]))
	{
		[self setSensitivity:(_sensitivity - 1)];
	}

	// Apply sensitivity
	// TODO Sensitivity scaling is basically broken.
	int accScaled;
	int j;
	unsigned int scaleOffset;
	if (_sensitivity)
	{
		for (i = 0; i < 3; i++)
		{
			accScaled = signals[i];
			for (j = 0, scaleOffset = (SIGNAL_RANGE >> 2); j++, scaleOffset >>= 1; j < _sensitivity)
			{
				accScaled -= scaleOffset;
			}
			accScaled << _sensitivity;
			signals[i] = (unsigned char) (accScaled >= SIGNAL_RANGE ? (SIGNAL_RANGE - 1) : (accScaled < 0 ? 0 : accScaled));
		}
	}

	if (_learnMode)
	{
		if (signals[SA_BUT_A] && signals[SA_BUT_B])
		{
			signals[SA_ACC_X] = _oldSignals[SA_ACC_X];
			signals[SA_ACC_Z] = _oldSignals[SA_ACC_Z];
		}
		else if (signals[SA_BUT_A])
		{
			signals[SA_ACC_X] = _oldSignals[SA_ACC_X];
			signals[SA_ACC_Y] = _oldSignals[SA_ACC_Y];
		}
		else if (signals[SA_BUT_B])
		{
			signals[SA_ACC_Y] = _oldSignals[SA_ACC_Y];
			signals[SA_ACC_Z] = _oldSignals[SA_ACC_Z];
		}
		else
		{
			signals[SA_ACC_X] = _oldSignals[SA_ACC_X];
			signals[SA_ACC_Y] = _oldSignals[SA_ACC_Y];
			signals[SA_ACC_Z] = _oldSignals[SA_ACC_Z];		
		}
	}

	Byte mData[256];
	Byte* mDataP = mData;
  
	for (i = 0; i < 3; i++)
	{
		if (signals[i] == _oldSignals[i]) continue;
		*mDataP++ = 0xB0 | _midiChannel;
		*mDataP++ = 0x10 + i;
		*mDataP++ = signals[i] >> 2; // MIDI resolution is only 128
	}
	
	Byte listStorage[256]; // avoid malloc
	MIDIPacketList *list = (MIDIPacketList *) listStorage;
	MIDIPacket *x = MIDIPacketListInit(list);
	x = MIDIPacketListAdd(list, sizeof(listStorage), x, MIDI_TIME_NOW, (mDataP - mData), mData);

	MIDISend(_midiOut, _midiDest, list);

	memcpy(_oldSignals, signals, SIGNAL_ARR_SIZE);
}

-(void) setLearnMode:(BOOL)learnMode
{
	// TODO notifications?
	NSLog(@"LearnMode %i", learnMode);
	_learnMode = learnMode;
}

-(void) setSensitivity:(int)sensitivity
{
	if (sensitivity > 3) sensitivity = 3;
	if (sensitivity < 0) sensitivity = 0;
	// TODO notifications?
	NSLog(@"Sensitivity set to %i", sensitivity);
	_sensitivity = sensitivity;
}

-(void) wiiRemoteDisconnected
{
	[_controller wiiRemoteDisconnected:_wiiId];
}

-(void) release
{
	[_wii closeConnection];
	[_wii release];
	[super release];
}

@end
